function _defineProperty(obj, key, value) { key = _toPropertyKey(key); if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }
function _toPropertyKey(arg) { var key = _toPrimitive(arg, "string"); return typeof key === "symbol" ? key : String(key); }
function _toPrimitive(input, hint) { if (typeof input !== "object" || input === null) return input; var prim = input[Symbol.toPrimitive]; if (prim !== undefined) { var res = prim.call(input, hint || "default"); if (typeof res !== "object") return res; throw new TypeError("@@toPrimitive must return a primitive value."); } return (hint === "string" ? String : Number)(input); }
import { subtract, distance, length, cosineSimilarity, equals, frechetDist, normalizeCurve, rotate } from '../geometry';
function average(arr) {
  const sum = arr.reduce((acc, val) => val + acc, 0);
  return sum / arr.length;
}

// All makemeahanzi characters have the same bounding box
const CHARACTER_BOUNDS = [{
  x: 0,
  y: -124
}, {
  x: 1024,
  y: 900
}];
const [from, to] = CHARACTER_BOUNDS;
const preScaledWidth = to.x - from.x;
const preScaledHeight = to.y - from.y;
export class Positioner {
  constructor(options) {
    _defineProperty(this, "padding", void 0);
    _defineProperty(this, "width", void 0);
    _defineProperty(this, "height", void 0);
    _defineProperty(this, "xOffset", void 0);
    _defineProperty(this, "yOffset", void 0);
    _defineProperty(this, "scale", void 0);
    const {
      padding,
      width,
      height
    } = options;
    this.padding = padding;
    this.width = width;
    this.height = height;
    const effectiveWidth = width - 2 * padding;
    const effectiveHeight = height - 2 * padding;
    const scaleX = effectiveWidth / preScaledWidth;
    const scaleY = effectiveHeight / preScaledHeight;
    this.scale = Math.min(scaleX, scaleY);
    const xCenteringBuffer = padding + (effectiveWidth - this.scale * preScaledWidth) / 2;
    const yCenteringBuffer = padding + (effectiveHeight - this.scale * preScaledHeight) / 2;
    this.xOffset = -1 * from.x * this.scale + xCenteringBuffer;
    this.yOffset = -1 * from.y * this.scale + yCenteringBuffer;
  }
  convertExternalPoint(point) {
    const x = (point.x - this.xOffset) / this.scale;
    const y = (this.height - this.yOffset - point.y) / this.scale;
    return {
      x,
      y
    };
  }
}
export class UserStroke {
  constructor(id, startingPoint, startingExternalPoint) {
    _defineProperty(this, "id", void 0);
    _defineProperty(this, "points", void 0);
    _defineProperty(this, "externalPoints", void 0);
    this.id = id;
    this.points = [startingPoint];
    this.externalPoints = [startingExternalPoint];
  }
  appendPoint(point, externalPoint) {
    this.points.push(point);
    this.externalPoints.push(externalPoint);
  }
}
export class Stroke {
  constructor(path, points, strokeNum) {
    let isInRadical = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : false;
    _defineProperty(this, "path", void 0);
    _defineProperty(this, "points", void 0);
    _defineProperty(this, "strokeNum", void 0);
    _defineProperty(this, "isInRadical", void 0);
    this.path = path;
    this.points = points;
    this.strokeNum = strokeNum;
    this.isInRadical = isInRadical;
  }
  getStartingPoint() {
    return this.points[0];
  }
  getEndingPoint() {
    return this.points[this.points.length - 1];
  }
  getLength() {
    return length(this.points);
  }
  getVectors() {
    let lastPoint = this.points[0];
    const pointsSansFirst = this.points.slice(1);
    return pointsSansFirst.map(point => {
      const vector = subtract(point, lastPoint);
      lastPoint = point;
      return vector;
    });
  }
  getDistance(point) {
    const distances = this.points.map(strokePoint => distance(strokePoint, point));
    return Math.min(...distances);
  }
  getAverageDistance(points) {
    const totalDist = points.reduce((acc, point) => acc + this.getDistance(point), 0);
    return totalDist / points.length;
  }
}
function generateStrokes(_ref) {
  let {
    radStrokes,
    strokes,
    medians
  } = _ref;
  const isInRadical = strokeNum => ((radStrokes === null || radStrokes === void 0 ? void 0 : radStrokes.indexOf(strokeNum)) ?? -1) >= 0;
  return strokes.map((path, index) => {
    const points = medians[index].map(pointData => {
      const [x, y] = pointData;
      return {
        x,
        y
      };
    });
    return new Stroke(path, points, index, isInRadical(index));
  });
}
export class Character {
  constructor(symbol, strokes) {
    _defineProperty(this, "symbol", void 0);
    _defineProperty(this, "strokes", void 0);
    this.symbol = symbol;
    this.strokes = strokes;
  }
}
export function parseCharData(symbol, charJson) {
  const strokes = generateStrokes(charJson);
  return new Character(symbol, strokes);
}
const AVG_DIST_THRESHOLD = 350; // bigger = more lenient
const COSINE_SIMILARITY_THRESHOLD = 0; // -1 to 1, smaller = more lenient
const START_AND_END_DIST_THRESHOLD = 250; // bigger = more lenient
const FRECHET_THRESHOLD = 0.4; // bigger = more lenient
const MIN_LEN_THRESHOLD = 0.35; // smaller = more lenient

export function strokeMatches(userStroke, character, strokeNum) {
  let options = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : {};
  const strokes = character.strokes;
  const points = stripDuplicates(userStroke.points);
  if (points.length < 2) {
    return {
      isMatch: false,
      meta: {
        isStrokeBackwards: false
      }
    };
  }
  const {
    isMatch: isMatchOne,
    meta,
    avgDist: avgDistOne
  } = getMatchData(points, strokes[strokeNum], options);
  if (!isMatchOne) {
    return {
      isMatch: isMatchOne,
      meta
    };
  }

  // if there is a better match among strokes the user hasn't drawn yet, the user probably drew the wrong stroke
  const laterStrokes = strokes.slice(strokeNum + 1);
  let closestMatchDist = avgDistOne;
  for (let i = 0; i < laterStrokes.length; i++) {
    const {
      isMatch: isMatchTwo,
      avgDist: avgDistTwo
    } = getMatchData(points, laterStrokes[i], {
      ...options,
      checkBackwards: false
    });
    if (isMatchTwo && avgDistTwo < closestMatchDist) {
      closestMatchDist = avgDistTwo;
    }
  }
  // if there's a better match, rather that returning false automatically, try reducing leniency instead
  // if leniency is already really high we can allow some similar strokes to pass
  if (closestMatchDist < avgDistOne) {
    // adjust leniency between 0.3 and 0.6 depending on how much of a better match the new match is
    const leniencyAdjustment = 0.6 * (closestMatchDist + avgDistOne) / (2 * avgDistOne);
    const {
      isMatch: isMatchThree,
      meta
    } = getMatchData(points, strokes[strokeNum], {
      ...options,
      leniency: (options.leniency || 1) * leniencyAdjustment
    });
    return {
      isMatch: isMatchThree,
      meta
    };
  }
  return {
    isMatch: isMatchOne,
    meta
  };
}
const startAndEndMatches = (points, closestStroke, leniency) => {
  const startingDist = distance(closestStroke.getStartingPoint(), points[0]);
  const endingDist = distance(closestStroke.getEndingPoint(), points[points.length - 1]);
  return startingDist <= START_AND_END_DIST_THRESHOLD * leniency && endingDist <= START_AND_END_DIST_THRESHOLD * leniency;
};

// returns a list of the direction of all segments in the line connecting the points
const getEdgeVectors = points => {
  const vectors = [];
  let lastPoint = points[0];
  points.slice(1).forEach(point => {
    vectors.push(subtract(point, lastPoint));
    lastPoint = point;
  });
  return vectors;
};
const directionMatches = (points, stroke) => {
  const edgeVectors = getEdgeVectors(points);
  const strokeVectors = stroke.getVectors();
  const similarities = edgeVectors.map(edgeVector => {
    const strokeSimilarities = strokeVectors.map(strokeVector => cosineSimilarity(strokeVector, edgeVector));
    return Math.max(...strokeSimilarities);
  });
  const avgSimilarity = average(similarities);
  return avgSimilarity > COSINE_SIMILARITY_THRESHOLD;
};
const lengthMatches = (points, stroke, leniency) => {
  return leniency * (length(points) + 25) / (stroke.getLength() + 25) >= MIN_LEN_THRESHOLD;
};
const stripDuplicates = points => {
  if (points.length < 2) return points;
  const [firstPoint, ...rest] = points;
  const dedupedPoints = [firstPoint];
  for (const point of rest) {
    if (!equals(point, dedupedPoints[dedupedPoints.length - 1])) {
      dedupedPoints.push(point);
    }
  }
  return dedupedPoints;
};
const SHAPE_FIT_ROTATIONS = [Math.PI / 16, Math.PI / 32, 0, -1 * Math.PI / 32, -1 * Math.PI / 16];
const shapeFit = (curve1, curve2, leniency) => {
  const normCurve1 = normalizeCurve(curve1);
  const normCurve2 = normalizeCurve(curve2);
  let minDist = Infinity;
  SHAPE_FIT_ROTATIONS.forEach(theta => {
    const dist = frechetDist(normCurve1, rotate(normCurve2, theta));
    if (dist < minDist) {
      minDist = dist;
    }
  });
  return minDist <= FRECHET_THRESHOLD * leniency;
};
const getMatchData = (points, stroke, options) => {
  const {
    leniency = 1,
    isOutlineVisible = false,
    checkBackwards = true
  } = options;
  const avgDist = stroke.getAverageDistance(points);
  const distMod = isOutlineVisible || stroke.strokeNum > 0 ? 0.5 : 1;
  const withinDistThresh = avgDist <= AVG_DIST_THRESHOLD * distMod * leniency;
  // short circuit for faster matching
  if (!withinDistThresh) {
    return {
      isMatch: false,
      avgDist,
      meta: {
        isStrokeBackwards: false
      }
    };
  }
  const startAndEndMatch = startAndEndMatches(points, stroke, leniency);
  const directionMatch = directionMatches(points, stroke);
  const shapeMatch = shapeFit(points, stroke.points, leniency);
  const lengthMatch = lengthMatches(points, stroke, leniency);
  const isMatch = withinDistThresh && startAndEndMatch && directionMatch && shapeMatch && lengthMatch;
  if (checkBackwards && !isMatch) {
    const backwardsMatchData = getMatchData([...points].reverse(), stroke, {
      ...options,
      checkBackwards: false
    });
    if (backwardsMatchData.isMatch) {
      return {
        isMatch,
        avgDist,
        meta: {
          isStrokeBackwards: true
        }
      };
    }
  }
  return {
    isMatch,
    avgDist,
    meta: {
      isStrokeBackwards: false
    }
  };
};
//# sourceMappingURL=index.js.map